//
// C++ Implementation: FlightDeckWindow
//
// Description: Flight Deck Windows, Altitude Dialog
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2011, 2012, 2013
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"
#include "sysdepsgui.h"
#include <iomanip>

#include "flightdeckwindow.h"

void FlightDeckWindow::fplatmosphere_dialog_open(void)
{
	if (!m_fplatmodialog || !m_sensors)
		return;
	m_fplatmodialog->open(m_sensors);
}

const unsigned int FlightDeckWindow::FPLAtmosphereDialog::nrkeybuttons;
const struct FlightDeckWindow::FPLAtmosphereDialog::keybutton FlightDeckWindow::FPLAtmosphereDialog::keybuttons[nrkeybuttons] = {
	{ "fplatmobutton0", GDK_KEY_KP_0, 90, 1 },
	{ "fplatmobutton1", GDK_KEY_KP_1, 87, 1 },
	{ "fplatmobutton2", GDK_KEY_KP_2, 88, 1 },
	{ "fplatmobutton3", GDK_KEY_KP_3, 89, 1 },
	{ "fplatmobutton4", GDK_KEY_KP_4, 83, 1 },
	{ "fplatmobutton5", GDK_KEY_KP_5, 84, 1 },
	{ "fplatmobutton6", GDK_KEY_KP_6, 85, 1 },
	{ "fplatmobutton7", GDK_KEY_KP_7, 79, 1 },
	{ "fplatmobutton8", GDK_KEY_KP_8, 80, 1 },
	{ "fplatmobutton9", GDK_KEY_KP_9, 81, 1 },
	{ "fplatmobuttondot", GDK_KEY_KP_Decimal, 91, 1 }
};

FlightDeckWindow::FPLAtmosphereDialog::FPLAtmosphereDialog(BaseObjectType *castitem, const Glib::RefPtr<Builder>& refxml)
	: Gtk::Window(castitem), m_calendaroffblock(0), m_spinbuttonoffblockh(0), m_spinbuttonoffblockm(0), m_spinbuttonoffblocks(0), 
	  m_spinbuttontaxitimem(0), m_spinbuttontaxitimes(0), m_spinbuttoncruisealt(0), m_spinbuttonwinddir(0), m_spinbuttonwindspeed(0), 
	  m_spinbuttonisaoffs(0), m_spinbuttonqff(0), m_buttontoday(0), m_buttondone(0), m_buttonsetwind(0), m_buttonsetisaoffs(0),
	  m_buttonsetqff(0), m_buttongetnwx(0), m_buttongfs(0), m_entryflighttime(0), m_entryfuel(0), m_sensors(0)
{
#ifndef GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
        signal_delete_event().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::on_delete_event));
#endif //GLIBMM_DEFAULT_SIGNAL_HANDLERS_ENABLED
	get_widget(refxml, "fplatmocalendaroffblock", m_calendaroffblock);
	get_widget(refxml, "fplatmospinbuttonoffblockh", m_spinbuttonoffblockh);
	get_widget(refxml, "fplatmospinbuttonoffblockm", m_spinbuttonoffblockm);
	get_widget(refxml, "fplatmospinbuttonoffblocks", m_spinbuttonoffblocks);
	get_widget(refxml, "fplatmospinbuttontaxitimem", m_spinbuttontaxitimem);
	get_widget(refxml, "fplatmospinbuttontaxitimes", m_spinbuttontaxitimes);
	get_widget(refxml, "fplatmospinbuttoncruisealt", m_spinbuttoncruisealt);
	get_widget(refxml, "fplatmospinbuttonwinddir", m_spinbuttonwinddir);
	get_widget(refxml, "fplatmospinbuttonwindspeed", m_spinbuttonwindspeed);
	get_widget(refxml, "fplatmospinbuttonisaoffs", m_spinbuttonisaoffs);
	get_widget(refxml, "fplatmospinbuttonqff", m_spinbuttonqff);
	get_widget(refxml, "fplatmobuttontoday", m_buttontoday);
	get_widget(refxml, "fplatmobuttondone", m_buttondone);
	get_widget(refxml, "fplatmobuttonsetwind", m_buttonsetwind);
	get_widget(refxml, "fplatmobuttonsetisaoffs", m_buttonsetisaoffs);
	get_widget(refxml, "fplatmobuttonsetqff", m_buttonsetqff);
	get_widget(refxml, "fplatmobuttongetnwx", m_buttongetnwx);
	get_widget(refxml, "fplatmobuttongfs", m_buttongfs);
	get_widget(refxml, "fplatmoentryflighttime", m_entryflighttime);
	get_widget(refxml, "fplatmoentryfuel", m_entryfuel);

        if (m_calendaroffblock)
                m_connoffblock = m_calendaroffblock->signal_day_selected().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::deptime_changed));
	if (m_spinbuttonoffblockh)
		m_connoffblockh = m_spinbuttonoffblockh->signal_value_changed().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::deptime_changed));
	if (m_spinbuttonoffblockm)
		m_connoffblockm = m_spinbuttonoffblockm->signal_value_changed().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::deptime_changed));
	if (m_spinbuttonoffblocks)
		m_connoffblocks = m_spinbuttonoffblocks->signal_value_changed().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::deptime_changed));
	if (m_spinbuttontaxitimem)
		m_conntaxitimem = m_spinbuttontaxitimem->signal_value_changed().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::deptime_changed));
	if (m_spinbuttontaxitimes)
		m_conntaxitimes = m_spinbuttontaxitimes->signal_value_changed().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::deptime_changed));
	if (m_buttontoday)
		m_buttontoday->signal_clicked().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::today_clicked));
	if (m_buttondone)
		m_buttondone->signal_clicked().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::done_clicked));
	if (m_buttonsetwind)
		m_buttonsetwind->signal_clicked().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::setwind_clicked));
	if (m_buttonsetisaoffs)
		m_buttonsetisaoffs->signal_clicked().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::setisaoffs_clicked));
	if (m_buttonsetqff)
		m_buttonsetqff->signal_clicked().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::setqff_clicked));
	if (m_buttongetnwx)
		m_buttongetnwx->signal_clicked().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::getnwx_clicked));
	if (m_buttongfs)
		m_buttongfs->signal_clicked().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::gfs_clicked));
	if (m_spinbuttoncruisealt)
		m_conncruisealt = m_spinbuttoncruisealt->signal_value_changed().connect(sigc::mem_fun(*this, &FPLAtmosphereDialog::cruisealt_changed));
	for (unsigned int i = 0; i < nrkeybuttons; ++i) {
		m_keybutton[i] = 0;
		get_widget(refxml, keybuttons[i].widget_name, m_keybutton[i]);
		if (!m_keybutton[i])
			continue;
		m_connkey[i] = m_keybutton[i]->signal_clicked().connect(sigc::bind(sigc::mem_fun(*this, &FPLAtmosphereDialog::key_clicked), i));
		m_keybutton[i]->set_can_focus(false);
	}
}

FlightDeckWindow::FPLAtmosphereDialog::~FPLAtmosphereDialog()
{
}

void FlightDeckWindow::FPLAtmosphereDialog::open(Sensors *sensors)
{
	m_sensors = sensors;
	if (m_sensors) {
		sensors_change(Sensors::change_all);
		show();
	} else {
		hide();
	}
}

void FlightDeckWindow::FPLAtmosphereDialog::sensors_change(Sensors::change_t changemask)
{
	if (!m_sensors || !changemask)
		return;
	if (changemask & Sensors::change_fplan) {
		const FPlanRoute& fpl(m_sensors->get_route());
		if (!fpl.get_nrwpt()) {
			m_sensors = 0;
			hide();
		}
		// set times
		m_connoffblock.block();
		m_connoffblockh.block();
		m_connoffblockm.block();
		m_connoffblocks.block();
		m_conntaxitimem.block();
		m_conntaxitimes.block();
		{
			time_t t(fpl.get_time_offblock_unix());
			struct tm utm;
			gmtime_r(&t, &utm);
			if (m_calendaroffblock) {
				m_calendaroffblock->select_month(utm.tm_mon, utm.tm_year + 1900);
				m_calendaroffblock->select_day(utm.tm_mday);
				m_calendaroffblock->mark_day(utm.tm_mday);
			}
			if (m_spinbuttonoffblockh)
				m_spinbuttonoffblockh->set_value(utm.tm_hour);
			if (m_spinbuttonoffblockm)
				m_spinbuttonoffblockm->set_value(utm.tm_min);
			if (m_spinbuttonoffblocks)
				m_spinbuttonoffblocks->set_value(utm.tm_sec);
			t = fpl[0].get_time_unix() - t;
			if (m_spinbuttontaxitimem)
				m_spinbuttontaxitimem->set_value(t / 60U);
			if (m_spinbuttontaxitimes)
				m_spinbuttontaxitimes->set_value(t % 60U);
		}
		m_connoffblock.unblock();
		m_connoffblockh.unblock();
		m_connoffblockm.unblock();
		m_connoffblocks.unblock();
		m_conntaxitimem.unblock();
		m_conntaxitimes.unblock();
		// find altitudes, average atmosphere
		{
			typedef std::set<int> altset_t;
			altset_t altset;
			float windx(0), windy(0), isaoffs(0), qff(0);
			double fuel(0);
			for (unsigned int i = 0; i < fpl.get_nrwpt(); ++i) {
				const FPlanWaypoint& wpt(fpl[i]);
				altset.insert(wpt.get_altitude());
				float x(0), y(0);
				sincosf(wpt.get_winddir_rad(), &y, &x);
				windx += x * wpt.get_windspeed_kts();
				windy += y * wpt.get_windspeed_kts();
				isaoffs += wpt.get_isaoffset_kelvin();
				qff += wpt.get_qff_hpa();
				fuel = wpt.get_fuel_usg();
			}
			{
				float s(1.f / fpl.get_nrwpt());
				windx *= s;
				windy *= s;
				isaoffs *= s;
				qff *= s;
			}
			float windspeed(sqrtf(windx * windx + windy * windy)), winddir(0.f);
			if (windspeed >= 0.001f)
				winddir = atan2f(windy, windx) * (float)(180.f / M_PI);
			m_conncruisealt.block();
			if (m_spinbuttoncruisealt) {
				m_spinbuttoncruisealt->set_sensitive(false);
				altset_t::const_reverse_iterator ai(altset.rbegin()), ae(altset.rend());
				if (ai != ae) {
					int crsalt(*ai++);
					m_spinbuttoncruisealt->set_value(crsalt);
					if (ai != ae) {
						double rmin, rmax;
						m_spinbuttoncruisealt->get_range(rmin, rmax);
						rmin = std::min(crsalt, *ai + 100);
						m_spinbuttoncruisealt->set_range(rmin, rmax);
						m_spinbuttoncruisealt->set_sensitive(true);
					}
				}						
			}
			m_conncruisealt.unblock();
			if (m_spinbuttonwinddir)
				m_spinbuttonwinddir->set_value(winddir);
			if (m_spinbuttonwindspeed)
				m_spinbuttonwindspeed->set_value(windspeed);
			if (m_spinbuttonisaoffs)
				m_spinbuttonisaoffs->set_value(isaoffs);
			if (m_spinbuttonqff)
				m_spinbuttonqff->set_value(qff);
			if (m_entryfuel)
				m_entryfuel->set_text(Glib::ustring::format(std::fixed, std::setprecision(1), fuel));
		}
		// compute total flight time
		if (m_entryflighttime) {
			time_t t(fpl[fpl.get_nrwpt() - 1].get_time_unix() - fpl[0].get_time_unix());
			m_entryflighttime->set_text(Conversions::time_str(t));
		}
	}
}

void FlightDeckWindow::FPLAtmosphereDialog::today_clicked(void)
{
	if (!m_calendaroffblock)
		return;
	Glib::Date d;
	d.set_time_current();
	m_calendaroffblock->select_month(d.get_month() - 1, d.get_year());
	m_calendaroffblock->select_day(d.get_day());
	deptime_changed();
}

void FlightDeckWindow::FPLAtmosphereDialog::done_clicked(void)
{
	hide();
	m_sensors = 0;
}

void FlightDeckWindow::FPLAtmosphereDialog::setwind_clicked(void)
{
	if (!m_sensors)
		return;
	FPlanRoute& fpl(m_sensors->get_route());
	if (!fpl.get_nrwpt())
		return;
	double winddir(0), windspeed(0);
	if (m_spinbuttonwinddir)
		winddir = m_spinbuttonwinddir->get_value();
	if (m_spinbuttonwindspeed)
		windspeed = m_spinbuttonwindspeed->get_value();
	fpl.set_winddir_deg(winddir);
	fpl.set_windspeed_kts(windspeed);
	deptime_changed(); // force recompute of flight plan
}

void FlightDeckWindow::FPLAtmosphereDialog::setisaoffs_clicked(void)
{
	if (!m_sensors)
		return;
	FPlanRoute& fpl(m_sensors->get_route());
	if (!fpl.get_nrwpt())
		return;
	double isaoffs(0);
	if (m_spinbuttonisaoffs)
		isaoffs = m_spinbuttonisaoffs->get_value();
	fpl.set_isaoffset_kelvin(isaoffs);
	deptime_changed(); // force recompute of flight plan
}

void FlightDeckWindow::FPLAtmosphereDialog::setqff_clicked(void)
{
	if (!m_sensors)
		return;
	FPlanRoute& fpl(m_sensors->get_route());
	if (!fpl.get_nrwpt())
		return;
	double qff(IcaoAtmosphere::std_sealevel_pressure);
	if (m_spinbuttonqff)
		qff = m_spinbuttonqff->get_value();
	fpl.set_qff_hpa(qff);
	deptime_changed(); // force recompute of flight plan
}

void FlightDeckWindow::FPLAtmosphereDialog::getnwx_clicked(void)
{
	if (!m_sensors)
		return;
	deptime_changed(); // force recompute of flight plan
	m_sensors->get_route().nwxweather(m_nwxweather, sigc::mem_fun(*this, &FPLAtmosphereDialog::fetchwind_cb));
}

void FlightDeckWindow::FPLAtmosphereDialog::fetchwind_cb(bool fplmodif)
{
	if (fplmodif)
	 	deptime_changed(); // force recompute of flight plan
}

void FlightDeckWindow::FPLAtmosphereDialog::gfs_clicked(void)
{
	if (!m_sensors)
		return;
	FPlanRoute& fpl(m_sensors->get_route());
	deptime_changed(); // force recompute of flight plan
	fpl.set_winddir(0);
	fpl.set_windspeed(0);
	FPlanRoute::GFSResult r(fpl.gfs(m_sensors->get_engine()->get_grib2_db()));
	if (r.is_modified())
		deptime_changed(); // force recompute of flight plan
}

void FlightDeckWindow::FPLAtmosphereDialog::deptime_changed(void)
{
	if (!m_sensors)
		return;
	FPlanRoute& fpl(m_sensors->get_route());
	struct tm utm;
	{
		time_t t(fpl.get_time_offblock_unix());
		gmtime_r(&t, &utm);
	}
	if (m_calendaroffblock) {
		guint year, month, day;
		m_calendaroffblock->get_date(year, month, day);
		if (year) {
			utm.tm_year = year - 1900;
			utm.tm_mon = month;
		}
		if (day)
			utm.tm_mday = day;
	}
	if (m_spinbuttonoffblockh)
		utm.tm_hour = m_spinbuttonoffblockh->get_value();
	if (m_spinbuttonoffblockm)
		utm.tm_min = m_spinbuttonoffblockm->get_value();
	if (m_spinbuttonoffblocks)
		utm.tm_sec = m_spinbuttonoffblocks->get_value();
	unsigned int taxis(0U), taxim(5U);
	if (m_spinbuttontaxitimem)
		taxim = m_spinbuttontaxitimem->get_value();
	if (m_spinbuttontaxitimes)
		taxis = m_spinbuttontaxitimes->get_value();
	m_sensors->nav_fplan_setdeparture(timegm(&utm), taxim * 60U + taxis);
}

void FlightDeckWindow::FPLAtmosphereDialog::cruisealt_changed(void)
{
	if (!m_sensors)
		return;
	FPlanRoute& fpl(m_sensors->get_route());
	int curcrsalt(std::numeric_limits<int>::min());
	for (unsigned int i = 0; i < fpl.get_nrwpt(); ++i) {
		FPlanWaypoint& wpt(fpl[i]);
		curcrsalt = std::max(curcrsalt, static_cast<int>(wpt.get_altitude()));
	}
	int newcrsalt(curcrsalt);
	if (m_spinbuttoncruisealt)
		newcrsalt = m_spinbuttoncruisealt->get_value();
	for (unsigned int i = 0; i < fpl.get_nrwpt(); ++i) {
		FPlanWaypoint& wpt(fpl[i]);
		if (wpt.get_altitude() != curcrsalt)
			continue;
		wpt.set_altitude(newcrsalt);
	}
	deptime_changed(); // force recompute of flight plan
}

bool FlightDeckWindow::FPLAtmosphereDialog::on_delete_event(GdkEventAny* event)
{
        hide();
        return true;
}

void FlightDeckWindow::FPLAtmosphereDialog::key_clicked(unsigned int i)
{
 	if (i >= nrkeybuttons)
		return;
	Gtk::Widget *w(get_focus());
	if (!w)
		return;
	for (unsigned int cnt(keybuttons[i].count); cnt > 0; --cnt)
		KeyboardDialog::send_key(w, 0, keybuttons[i].key, keybuttons[i].hwkeycode);
}
